package es.usal.bisite.heatmap;

/*    
Copyright (C) 2012 http://software-talk.org/ (developer@software-talk.org)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.example.bluepin.R;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Point;
import android.graphics.drawable.BitmapDrawable;
import android.graphics.drawable.Drawable;


/**
* This class can create a heatmap from a given list of points.
* The output will be a picture visualising the occurences of points.
* <p>
* CIRCLEPIC (an image of a circle, from black to transparent)
* <p>
* and
* <p>
* SPECTRUMPIC (a color gradiant
*  where the color that represents "most" is at the bottom)
* <p>
* must be in the user.dir in:
* <p>
* heatmap/bolillaT.png
* <p>
* and
* <p>
* heatmap/colors.png
*
* @author sam
*/
public class HeatMap 
{


	private static final int HALFCIRCLEPICSIZE = 32;	

	private Map<Integer, List<Point>> map;	
	private int maxOccurance = 1;	
	private int maxXValue;	
	private int maxYValue;
	private Bitmap hMap;
	private Context context;

	/**
	 * constructs new instance of HeatMap from given list of points.
	 * Depending on the amount of points, this may take a while,
	 * as the points are being sorted.
	 *
	 * @param points the list of points
	 * @param lvlMap name of file to lay heatmap over
	 */
	//====================================================================
	public HeatMap(List<Point> points,Context _context) 
	//====================================================================
	{
		this.context=_context;
		Drawable blankDrawable = context.getResources().getDrawable(R.drawable.fisicas);
		Bitmap mMap=((BitmapDrawable)blankDrawable).getBitmap();
		initMap(points,mMap);
	}

	/**
	 * initiate map.
	 * counts and sorts points and figures out max x and y values as well
	 * as the maximal amount of points with same coordinates.
	 *	 max x and y values will be used for the size of the heatmap.
	 *
	 * @param points list of points
	 */
	//====================================================================
	private void initMap(List<Point> points,Bitmap mapPic) 
	//====================================================================
	{
		map = new HashMap<Integer, List<Point>>();
 
		maxXValue = mapPic.getWidth();
		maxYValue = mapPic.getHeight();

		int pointSize = points.size();
		for (int i = 0; i < pointSize; i++) 
		{
			Point point = points.get(i);
			// add point to right list.
			int hash = getkey(point);
			if (map.containsKey(hash)) 
			{
				List<Point> thisList = map.get(hash);
				thisList.add(point);
				if (thisList.size() > maxOccurance) 
				{
					maxOccurance = thisList.size();
				}
				// if list did not exist, create new one and add point.
			} 
			else 
			{
				List<Point> newList = new LinkedList<Point>();
				newList.add(point);
				map.put(hash, newList);
			}
		}
	}

	//====================================================================
	public Bitmap getResult(Bitmap base)
	//====================================================================
	{
		Drawable blankDrawable = context.getResources().getDrawable(R.drawable.fisicas);
		Bitmap mMap=((BitmapDrawable)blankDrawable).getBitmap();
		Canvas canvas = new Canvas(mMap);
		
		Paint p =new Paint();
		p.setAlpha(150);
		canvas.drawBitmap(hMap,0,0, p);
		
		return mMap;
	}
	/**
	 * creates the heatmap.
	 *
	 * @param multiplier calculated opacity of every point will be
	 * multiplied by this value. This leads to a HeatMap that is easier to read,
	 * especially when there are not too many points or the points are too
	 * spread out. Pass 1.0f for original.
	 */
	//====================================================================
	public void createHeatMap(float multiplier) 
	//====================================================================
	{
		Drawable blankDrawable = context.getResources().getDrawable(R.drawable.bolilla);
		Bitmap circle=((BitmapDrawable)blankDrawable).getBitmap();
		
		Bitmap.Config conf = Bitmap.Config.ARGB_8888; // see other conf types
		hMap = Bitmap.createBitmap(this.maxXValue, this.maxYValue, conf); // this creates a MUTABLE bitmap
		Canvas canvas = new Canvas(hMap);
		
		//Se pinta en blanco el bitmap
		canvas.drawRGB(255, 255, 255);
		Iterator<List<Point>> iterator = map.values().iterator();
		while (iterator.hasNext()) 
		{
			List<Point> currentPoints = iterator.next();

			// calculate opaqueness
			// based on number of occurences of current point
			float opaque = currentPoints.size() / (float) maxOccurance;

			// adjust opacity so the heatmap is easier to read
			opaque = opaque * multiplier;
			if (opaque > 1) 
			{
				opaque = 1;
			}

			Point currentPoint = currentPoints.get(0);

			// draw a circle which gets transparent from middle to outside
			// (which opaqueness is set to "opaque")
			// at the position specified by the center of the currentPoint
		  
			Paint p =new Paint();
			p.setAlpha((int)(opaque*255));
			canvas.drawBitmap(circle, (currentPoint.x - HALFCIRCLEPICSIZE), (currentPoint.y - HALFCIRCLEPICSIZE), p);
		}

		// negate the image
		hMap = negateImage(hMap);

		// remap black/white with color spectrum from white over red, orange,
		// yellow, green to blue
		remap(hMap);
		
		//Pintar
	}

	/**
 	* remaps black and white picture with colors.
 	* It uses the colors from SPECTRUMPIC. The whiter a pixel is, the more it
 	* will get a color from the bottom of it. Black will stay black.
 	*
 	* @param heatMapBW black and white heat map
 	*/
	//====================================================
	private void remap(Bitmap heatMapBW) 
	//====================================================
	{
		Drawable blankDrawable = context.getResources().getDrawable(R.drawable.colors);
		Bitmap colorGradiant=((BitmapDrawable)blankDrawable).getBitmap();

		int width = heatMapBW.getWidth();
		int height = heatMapBW.getHeight();
		int gradientHight = colorGradiant.getHeight() - 1;
		
		for (int i = 0; i < width; i++) 
		{
			for (int j = 0; j < height; j++) 
			{
            	// get heatMapBW color values:
            	int rGB = heatMapBW.getPixel(i, j);

            	// calculate multiplier to be applied to height of gradiant.
            	float multiplier = rGB & 0xff; // blue
            	multiplier *= ((rGB >>> 8)) & 0xff; // green
            	multiplier *= (rGB >>> 16) & 0xff; // red
            	multiplier /= 16581375; // 255f * 255f * 255f

            	// apply multiplier
            	int y = (int) (multiplier * gradientHight);

            	// remap values
            	// calculate new value based on whitenes of heatMap
            	// (the whiter, the more a color from the top of colorGradiant
            	// will be chosen.
            	int mapedRGB = colorGradiant.getPixel(0, y);
            	// set new value
            	heatMapBW.setPixel(i, j, mapedRGB);
        	}
    	}
	}

	/**
	 * returns a negated version of this image.
	 *
	 * @param img buffer to negate
	 * @return negated buffer
	 */
	//====================================================
	private Bitmap negateImage(Bitmap img) 
	//====================================================
	{
		int width = img.getWidth();
		int height = img.getHeight();
		
		for (int x = 0; x < width; x++) 
		{
			for (int y = 0; y < height; y++) 
			{
				int rGB = img.getPixel(x, y);

				//Swaps values
				//i.e. 255, 255, 255 (white)
				//becomes 0, 0, 0 (black)
				int r = Math.abs(((rGB >>> 16) & 0xff) - 255); // red inverted
				int g = Math.abs(((rGB >>> 8) & 0xff) - 255); // green inverted
				int b = Math.abs((rGB & 0xff) - 255); // blue inverted

				// transform back to pixel value and set it
				img.setPixel(x, y, (r << 16) | (g << 8) | b);
			}
    }
    return img;
}


/**
 * returns a hash calculated by the given point.
 *
 * @param p a point
 * @return hash value
 */
private int getkey(Point p) {
    return ((p.x << 19) | (p.y << 7));
}

}
